﻿using System;
using System.Collections;
using System.Configuration;
using System.Globalization;
using System.IO;
using System.Net;
using MailBee.Mime;
using System.Net.Sockets;
using System.Text;
using System.Text.RegularExpressions;

namespace WebMailLite
{
	public struct WMServerUser
	{
		public string Domain;
		public string Name;
		public string Password;
		public string Type;

		public static WMServerUser Parse(string src)
		{
			WMServerUser user = new WMServerUser();

			Regex r = new Regex(@"""(?<domain>[^""]*)""\t""(?<name>[^""]*)""\t""(?<password>[^""]*)""\t""(?<type>[^""]*)""");
			Match m = r.Match(src);
			if (m.Success)
			{
				user.Domain = m.Groups["domain"].Value;
				user.Name = m.Groups["name"].Value;
				user.Password = m.Groups["password"].Value;
				user.Type = m.Groups["type"].Value;
			}

			return user;
		}
	}

	public class FileDateComparer : IComparer
	{
		#region IComparer Members

		public int Compare(object x, object y)
		{
			if (x == y) return 0;
			FileInfo fix = x as FileInfo;
			FileInfo fiy = y as FileInfo;
			if (fix == null) return -1;
			if (fiy == null) return 1;

			return fix.CreationTimeUtc.CompareTo(fiy.CreationTimeUtc);
		}

		#endregion
	}

	/// <summary>
	/// Summary description for XMailStorage.
	/// </summary>
	public class WMServerStorage : MailServerStorage
	{
		protected TcpClient _tcpClient = null;
		protected NetworkStream _nwStream = null;
		protected string _hostname = string.Empty;
		protected int _port = 0;
		protected WebmailResourceManager _resMan = null;

		protected string _response = string.Empty;

		public WMServerStorage(string hostname, int port, Account acct) : base(acct)
		{
			_hostname = hostname;
			_port = port;

			_resMan = (new WebmailResourceManagerCreator()).CreateResourceManager();
		}

        public WMServerStorage(string hostname, Account acct)
            : this(hostname, WMServerPort, acct)
        {
        }

		public WMServerStorage(Account acct)
			: this(WMServerHost, WMServerPort, acct)
		{
		}

		public static string WMServerHost
		{
			get
			{
				WebmailSettings settings = (new WebMailSettingsCreator()).CreateWebMailSettings();
				return settings.WmServerHost;
			}
		}

		public static int WMServerPort
		{
			get
			{
				WebmailSettings settings = (new WebMailSettingsCreator()).CreateWebMailSettings();
				return settings.XMailControlPort;
			}
		}

		public void ConnectAndLoginAdmin()
		{
			try
			{

				WebmailSettings settings = (new WebMailSettingsCreator()).CreateWebMailSettings();
				_tcpClient = new TcpClient();
				_tcpClient.Connect(_hostname, _port);
				_nwStream = _tcpClient.GetStream();

				ReceiveResponse(); // receive hello from server

				SendRequest(string.Format("{0}	{1}\r\n", settings.XMailLogin, settings.XMailPass));

				ReceiveResponse();
			}
			catch(Exception)
			{
				throw new WebMailException(_resMan.GetString("ErrorPOP3Connect"));
			}
		}

		public override void DeleteMessages(string[] messageUidSet)
		{
			try
			{
				WmServerFS serverFS = new WmServerFS(_account.Email);
				serverFS.DeleteMessages(messageUidSet);
			}
			catch (Exception ex)
			{
				throw PrepareFriendlyException(ex, _resMan.GetString("PROC_CANT_DEL_MSGS"));
			}
		}

		protected virtual void DisconnectAdmin()
		{
			if (_nwStream != null) _nwStream.Close();
			if (_tcpClient != null) _tcpClient.Close();
		}

		public override int GetInboxMessageCount(string fullPath)
		{
			return GetMessageCount();
		}

		public override int GetInboxUnreadMessageCount(string fullPath)
		{
			return 0;
		}

		public override bool IsConnected()
		{
			return base.IsConnected();
		}

		public override long GetMailStorageSize()
		{
			try
			{
				WmServerFS serverFS = new WmServerFS(_account.Email);
				return serverFS.GetMailStorageSize();
			}
			catch (Exception ex)
			{
				throw PrepareFriendlyException(ex, _resMan.GetString("PROC_CANT_MAIL_SIZE"));
			}
		}

		public override int GetMessageCount()
		{
			try
			{
				WmServerFS serverFS = new WmServerFS(_account.Email);
				return serverFS.GetMessageCount();
			}
			catch (Exception ex)
			{
				throw PrepareFriendlyException(ex, _resMan.GetString("PROC_CANT_GET_MESSAGES_COUNT"));
			}
		}

		public override MailMessage LoadMessage(string uid)
		{
			return LoadMessage(uid, false);
		}

		protected MailMessage LoadMessage(string uid, bool headersOnly)
		{
			try
			{
				MailMessage message = null;
				message = new XMailMessage();
				((XMailMessage)message).UidXMail = uid;
				message.Parser.ParseHeaderOnly = headersOnly;

				WmServerFS serverFS = new WmServerFS(_account.Email);
				string path = serverFS.GetMessagePath(uid);

				message.LoadMessage(path);
				if (message is XMailMessage)
				{
					((XMailMessage)message).SizeXMail = (new FileInfo(path)).Length;
				}

				return message;
			}
			catch (Exception ex)
			{
				throw PrepareFriendlyException(ex, _resMan.GetString("PROC_MSG_HAS_DELETED"));
			}
		}

		public override MailMessageCollection LoadMessageHeaders(int pageNumber)
		{
			try
			{
				WmServerFS serverFS = new WmServerFS(_account.Email);
				serverFS.MoveMessagesFromNewToCur();

				MailMessageCollection messages = new MailMessageCollection();

				FileInfo[] fis = serverFS.GetMessagesFileInfo();
				Array.Sort(fis, new FileDateComparer());

				WebmailSettings settings = (new WebMailSettingsCreator()).CreateWebMailSettings();
				int msgPerPage = settings.MailsPerPage;
				int msgCount = fis.Length;

				int startIndex = 0;
				int length = msgPerPage;
				if ((pageNumber * msgPerPage) > msgCount)
				{
					if (msgCount > msgPerPage)
					{
						length = msgCount % msgPerPage;
					}
					else
					{
						length = msgCount;
					}
				}
				else
				{
					startIndex = (msgCount - (pageNumber * msgPerPage));
				}
				for (int i = 0; i < length; i++)
				{
					if (((startIndex + i) < msgCount) && ((startIndex + i) >= 0))
					{
						messages.Add(LoadMessage(fis[startIndex + i].Name, true));
					}
				}

				CollectionBase cb = messages as CollectionBase;
				if (cb != null)
				{
					Utils.ReverseCollection(ref cb);
				}
				return messages;
			}
			catch (Exception ex)
			{
				throw PrepareFriendlyException(ex, _resMan.GetString("PROC_CANT_GET_MSG_LIST"));
			}
		}

		#region Admin Functions

		public virtual string[] GetDomainList()
		{
			try
			{
				ExecuteAdminCommand(string.Format("domainlist"));
			}
			catch (WebMailWMServerException ex)
			{
				throw PrepareFriendlyException(ex, "Retrieve domain list failed.");
			}
			if (_response != null)
			{
				string[] strs = PrepareResultList();
				ArrayList list = new ArrayList();
				foreach (string str in strs)
				{
					list.Add(str.Trim(new char[] { '"' }));
				}
				return (string[])list.ToArray(typeof(string));
			}
			return new string[0];
		}

		public virtual bool IsDomainExists(string domain)
		{
			string[] domains = GetDomainList();
			foreach (string d in domains)
			{
				if (string.Compare(domain, d, true, CultureInfo.InvariantCulture) == 0)
				{
					return true;
				}
			}
			return false;
		}

		public virtual WMServerUser[] GetUserList()
		{
			return GetUserList(null, null);
		}

		public virtual WMServerUser[] GetUserList(string domain, string user)
		{
			try
			{
				string command = "userlist";
				if (((domain != null) && (domain.Length > 0)) && ((user != null) && (user.Length > 0)))
				{
					command = string.Format(@"userlist	{0}	{1}", domain, user);
				}
				else if ((domain != null) && (domain.Length > 0))
				{
					command = string.Format(@"userlist	{0}", domain);
				}
				ExecuteAdminCommand(command);
			}
			catch (WebMailWMServerException ex)
			{
				throw PrepareFriendlyException(ex, "Retrieve accounts list failed.");
			}
			if (_response != null)
			{
				string[] strs = PrepareResultList();
				ArrayList list = new ArrayList();
				foreach (string str in strs)
				{
					list.Add(WMServerUser.Parse(str));
				}
				return (WMServerUser[])list.ToArray(typeof(WMServerUser));
			}
			return new WMServerUser[0];
		}

		public virtual bool IsUserExists(string user)
		{
			return IsUserExists(null, user);
		}

		public virtual bool IsUserExists(string domain, string user)
		{
			WMServerUser[] users = GetUserList(domain, null);
			if ((users != null) && (users.Length > 0))
			{
				foreach (WMServerUser u in users)
				{
					if (string.Compare(u.Name, user, true, CultureInfo.InvariantCulture) == 0)
					{
						return true;
					}
				}
			}
			return false;
		}

		#endregion

		protected void SendRequest(string command)
		{
			if (_nwStream == null) throw new WebMailWMServerException(_resMan.GetString("PROC_CANT_GET_MSG_LIST"));

			byte[] sendBytes = Encoding.UTF8.GetBytes(command);
			_nwStream.Write(sendBytes, 0, sendBytes.Length);
		}

		protected string ReceiveResponse()
		{
			if (_nwStream == null) throw new WebMailWMServerException(_resMan.GetString("PROC_CANT_GET_MSG_LIST"));

			bool multiline = false;
			StreamReader sr = new StreamReader(_nwStream, true);
			_response = sr.ReadLine();
			if (!IsPositiveResponse(_response, out multiline))
			{
				throw new WebMailWMServerException(_response);
			}
			if (multiline)
			{
				StringBuilder sb = new StringBuilder();
				string str = string.Empty;
				do
				{
					str = sr.ReadLine();
					sb.Append(str + "\r\n");
				}
				while (str != ".");
				_response = sb.ToString();
			}
			return _response;
		}

		protected bool IsPositiveResponse(string response, out bool multiline)
		{
			multiline = false;
			if ((response != null) && (response.Length > 0))
			{
				if (response.Length >= 6)
				{
					multiline = (string.Compare(response.Substring(1, 5), "00100", true, CultureInfo.InvariantCulture) == 0) ? true : false;
				}
				return (response[0] == '+') ? true : false;
			}
			return false;
		}

		protected void ExecuteAdminCommand(string command)
		{
			if (command == null) throw new WebMailWMServerException(Constants.wmServCmdNull);

			if (!command.EndsWith("\r\n")) command += "\r\n";

			try
			{
				ConnectAndLoginAdmin();

				SendRequest(command);

				ReceiveResponse();

				SendRequest("quit\r\n");

				//ReceiveResponse(); // comment because output results will be overwrite with OK by QUIT
			}
			catch (WebMailException)
			{
				throw;
			}
			catch (Exception ex)
			{
				throw new WebMailWMServerException(ex);
			}
			finally
			{
				DisconnectAdmin();
			}
		}

		protected WebMailWMServerException PrepareFriendlyException(Exception ex, string friendlyMessage)
		{
			string message = string.Format(@"{0}

Server response: {1}", friendlyMessage, ex.Message);
			return new WebMailWMServerException(message);
		}

		private string[] PrepareResultList()
		{
			ArrayList result = Utils.Split(_response, "\r\n");

			string lastStr = (string)result[result.Count - 1];
			if (string.Compare(lastStr, ".", true, CultureInfo.InvariantCulture) == 0) result.RemoveAt(result.Count - 1);

			return (string[])result.ToArray(typeof(string));
		}

		public static string EncryptPassword(string password)
		{
			StringBuilder sb = new StringBuilder();
			byte[] bytes = Encoding.UTF8.GetBytes(password);
			for (int i = 0; i < bytes.Length; i++)
			{
				byte b = (byte)((bytes[i] ^ 101) & 0xFF);
				sb.Append(b.ToString("X2"));
			}
			return sb.ToString();
		}

		public static string DecryptPassword(string password)
		{
			if (password == null) return string.Empty;
			string result = string.Empty;
			if ((password.Length > 0) && (password.Length % 2 == 0))
			{
				byte[] decryptedBytes = new byte[password.Length / 2];
				int startIndex = 0;
				int index = 0;
				while (startIndex < password.Length)
				{
					string strByte = password.Substring(startIndex, 2);
					int b = 0;
					if (int.TryParse(strByte, NumberStyles.HexNumber, CultureInfo.InvariantCulture, out b))
					{
						b = ((b & 0xFF) ^ 101);
						decryptedBytes[index] = (byte)b;
					}
					startIndex += 2;
					index++;
				}
				result = Encoding.UTF8.GetString(decryptedBytes);
			}
			return result;
		}
	}

	public class XMailMessage : MailMessage
	{
		private string _uid = string.Empty;
		public string UidXMail
		{
			get { return _uid; }
			set { _uid = value; }
		}

		private long _size = 0;
		public long SizeXMail
		{
			get { return _size; }
			set { _size = value; }
		}
	}

}
